@@ -164,7 +164,7 @@ bundle_data_new (GFile  *file,
 
   data->file = g_object_ref (file);
   if (gpg_data)
-    data->gpg_data = g_object_ref (gpg_data);
+    data->gpg_data = g_bytes_ref (gpg_data);
 
   return data;
 }
@@ -1621,6 +1621,10 @@ flatpak_transaction_add_ref (FlatpakTransaction             *self,
  *
  * Adds installing the given ref to this transaction.
  *
+ * The @remote can either be a configured remote of the installation,
+ * or a file:// uri pointing at a local repository to install from,
+ * in which case an origin remote is created.
+ *
  * Returns: %TRUE on success; %FALSE with @error set on failure.
  */
 gboolean
@@ -2271,7 +2275,9 @@ handle_suggested_remote_name (FlatpakTransaction *self, GKeyFile *keyfile, GErro
   if (url == NULL)
     return TRUE;
 
-  collection_id = g_key_file_get_string (keyfile, FLATPAK_REF_GROUP, FLATPAK_REF_COLLECTION_ID_KEY, NULL);
+  collection_id = g_key_file_get_string (keyfile, FLATPAK_REF_GROUP, FLATPAK_REF_DEPLOY_COLLECTION_ID_KEY, NULL);
+  if (collection_id == NULL || *collection_id == '\0')
+    collection_id = g_key_file_get_string (keyfile, FLATPAK_REF_GROUP, FLATPAK_REF_COLLECTION_ID_KEY, NULL);
 
   if (remote_is_already_configured (self, url, collection_id))
     return TRUE;
